package org.belot.server;

import org.belot.protocol.*;
import java.util.*;

public class BelotGameImpl implements BelotGame {

  // Game management:

  // sorted players:
  private int firstTurn;

  // cards:
  private BelotCard[][] cardsPerPlayer = new BelotCard[4][BelotDeck.CARDS_PER_PLAYER];
  private Vector[] takenCardsPerTeam = new Vector[] {new Vector(32), new Vector(32)};
  private Announces[] announcesPerPlayer = new Announces[] {
    new Announces(), new Announces(), new Announces(), new Announces()};
  private int[] belotsPerPlayer = new int[4];
  private BelotDeck deck = new BelotDeck();
  private BelotCard[] handOfCards = new BelotCard[4];
  private int numberOfHand;

  // team score:
  private int[] teamScore = new int[2];
  private int hanging;

  // trump:
  private BelotColor trump;
  private int bettingTeam;
  private boolean contra;
  private boolean recontra;
  private boolean isInternal;
  private int lastWinner;
  private boolean lastWasNoCardsTaken = false;

  // turns:
  private int bettingTurn;
  private int lastBettingPlayer;
  private int playingTurn;

  public BelotGameImpl() {
    synchronized (this) {
      this.teamScore[0] = this.teamScore[1] = MAX_TEAM_SCORE;
    }
  }

  // METHODS:

  // Game:

  public synchronized void startGame() {
    // make lastWasNoCardsTaken false:
    this.lastWasNoCardsTaken = false;

    // null player cards:
    for (int index = 0; index < 4; ++index) {
      for (int i = 0; i < BelotDeck.CARDS_PER_PLAYER; ++i) {
        cardsPerPlayer[index][i] = null;
      }
    }

    // null the team scores:
    teamScore[0] = teamScore[1] = 0;
    hanging = 0;

    // fix the turns:
    firstTurn = (int)(Math.random() * 4);
    bettingTurn = -1;
    playingTurn = -1;
  }

  public synchronized boolean isGameFinished() {
    return
      teamScore[0] >= MAX_TEAM_SCORE ||
      teamScore[1] >= MAX_TEAM_SCORE;
  }

  public synchronized boolean getLastWasNoCardsTaken() {
    return this.lastWasNoCardsTaken;
  }

  public synchronized int getTeamScore(int team) {
    return teamScore[team];
  }

  // Dealing:
  public synchronized void startDealing() {
    // make lastWasNoCardsTaken false:
    this.lastWasNoCardsTaken = false;

    // null the hand:
    numberOfHand = 0;
    handOfCards[0] = handOfCards[1] = handOfCards[2] = handOfCards[3] = null;

    // null the trump:
    trump = null;
    bettingTeam = -1;
    lastBettingPlayer = -1;
    contra = false;
    recontra = false;

    // remove the cards:
    takenCardsPerTeam[0].removeAllElements();
    takenCardsPerTeam[1].removeAllElements();

    announcesPerPlayer[0].clear();
    announcesPerPlayer[1].clear();
    announcesPerPlayer[2].clear();
    announcesPerPlayer[3].clear();

    belotsPerPlayer[0] = 0;
    belotsPerPlayer[1] = 0;
    belotsPerPlayer[2] = 0;
    belotsPerPlayer[3] = 0;

    // fix the turns:
    firstTurn = (firstTurn + 1) % 4;
    bettingTurn = firstTurn;
    playingTurn = firstTurn;

    // shuffle:
    deck.shuffle();

    // deal 3 + 2:
    for (int i = 0; i < 4; ++i) {
      deck.deal(cardsPerPlayer[(i + firstTurn) % 4], 0, 3);
    }
    for (int i = 0; i < 4; ++i) {
      deck.deal(cardsPerPlayer[(i + firstTurn) % 4], 3, 2);
    }
  }

  public synchronized void finishBetting() {
    for (int i = 0; i < 4; ++i) {
      deck.deal(cardsPerPlayer[(i + firstTurn) % 4], 5, 3);
    }
  }

  public synchronized boolean isDealingFinished() {
    return (trump == null) || (numberOfHand >= 8);
  }

  private static int round(int x) {
    int result = x / 10;
    if (x % 10 >= 5) {
      ++result;
    }
    return result;
  }

  private void addToResult(int[] result, int resultIndex, Vector vec, int lastIndex) {
    for (int i = 0; i <= lastIndex; ++i) {
      int[] announce = (int[])vec.elementAt(i);
      result[resultIndex] += Announces.getScore(announce[0]);
    }
  }

  private void sortAnnounces(Vector vec) {
    HeapSort.heapSort(vec, announcesPerPlayer[0]);
  }

  private void addAnnounces(int[] result) {
    // 1. add belots:
    for (int i = 0; i < 4; ++i) {
      result[getTeamOfPlayer(i)] += belotsPerPlayer[i] * 20;
    }

    // 2. sort the announces:
    Vector team0 = new Vector();
    Vector team1 = new Vector();
    announcesPerPlayer[0].addToVector(team0);
    announcesPerPlayer[1].addToVector(team1);
    announcesPerPlayer[2].addToVector(team0);
    announcesPerPlayer[3].addToVector(team1);
    sortAnnounces(team0);
    sortAnnounces(team1);

    // 3. remove the top equal announces:
    int lastIndex0 = team0.size() - 1;
    int lastIndex1 = team1.size() - 1;
    Announces comparator = announcesPerPlayer[0];
    while (lastIndex0 >= 0 && lastIndex1 >= 0) {
      if (comparator.compare(team0.elementAt(lastIndex0), team1.elementAt(lastIndex1)) != 0) {
        break;
      }
      // if they are equal, move on backwards:
      while (lastIndex0 > 0 && comparator.compare(team0.elementAt(lastIndex0), team0.elementAt(lastIndex0 - 1)) == 0) {
        --lastIndex0;
      }
      while (lastIndex1 > 0 && comparator.compare(team1.elementAt(lastIndex1), team1.elementAt(lastIndex1 - 1)) == 0) {
        --lastIndex1;
      }
      --lastIndex0;
      --lastIndex1;
    }
    if (lastIndex0 >= 0 && lastIndex1 >= 0) {
      // add only to the team that has the highest one:
      int cmp = comparator.compare(team0.elementAt(lastIndex0), team1.elementAt(lastIndex1));
      if (cmp < 0) {
        // team1 win their announces:
        addToResult(result, 1, team1, lastIndex1);
      } else {
        // team0 win their announces:
        addToResult(result, 0, team0, lastIndex0);
      }
    } else {
      // one of these will not happen:
      addToResult(result, 0, team0, lastIndex0);
      addToResult(result, 1, team1, lastIndex1);
    }
  }

  public synchronized int[] calculateDealingScore() {
    int[] result = new int[2];

    // 1. calculate the cards score:
    for (int team = 0; team < 2; ++team) {
      result[team] = 0;
      int size = takenCardsPerTeam[team].size();
      for (int i = 0; i < size; ++i) {
        BelotCard card = (BelotCard)takenCardsPerTeam[team].elementAt(i);
        result[team] += card.getScore(trump);
      }
    }

    // 2. add last 10:
    result[getTeamOfPlayer(lastWinner)] += 10;

    // 3. double for NoTrump:
    if (trump == BelotColor.NoTrump) {
      result[0] *= 2;
      result[1] *= 2;
    }

    // 4. get the difference in rounding:
    int difference = calculateDifference(round(result[0]), round(result[1]), trump);

    // 5. add the announces:
    if (trump != BelotColor.NoTrump) {
      addAnnounces(result);
    }

    // 6. add the no taken cards penalty:
    this.lastWasNoCardsTaken = false;
    for (int i = 0; i < 2; ++i) {
      if (takenCardsPerTeam[i].size() == 0) {
        result[1 - i] += 90;
        contra = false;
        recontra = false;
        this.lastWasNoCardsTaken = true;
      }
    }

    // 7. see if the game is hanging:
    boolean isHanging = result[0] == result[1];

    System.out.println("result[0] = " + result[0]);
    System.out.println("result[1] = " + result[1]);

    // 8. calculate the internal score:
    int otherTeam = 1 - bettingTeam;
    isInternal = result[bettingTeam] < result[otherTeam];
    int winnerTeam;
    if (result[0] < result[1]) {
      winnerTeam = 1;
    } else if (result[0] == result[1]) {
      winnerTeam = -1;
    } else {
      winnerTeam = 0;
    }
    System.out.println("---------->>>>>>>>>>>> IS INTERNAL : " + isInternal);
    System.out.println("---------->>>>>>>>>>>> WINNER TEAM : " + winnerTeam);
    if (isInternal) {
      result[otherTeam] += result[bettingTeam];
      result[bettingTeam] = 0;
    } else if (contra || recontra) {
      result[winnerTeam] += result[1 - winnerTeam];
      result[1 - winnerTeam] = 0;
    }

    // 9. round again for the last time::
    int rounding0 = round(result[0]);
    int rounding1 = round(result[1]);

    // 10. see which result belongs to whom:
    if (isHanging) {
      int winner = otherTeam;
      result[1 - winner] = 0;
      result[winner] = rounding0;
      if (difference < 0) {
        result[winner] += difference;
      } else {
        hanging += difference;
      }
      hanging += rounding0;
      teamScore[winner] += result[winner];
    } else {
      int winner = result[0] > result[1] ? 0 : 1;
      result[winner]     = (winner == 0 ? rounding0 : rounding1);
      result[1 - winner] = (winner == 1 ? rounding0 : rounding1);
      if (difference < 0) {
        result[winner] += difference;
      } else {
        result[1 - winner] += difference;
      }
      teamScore[0] += result[0];
      teamScore[1] += result[1];
      teamScore[winner] += hanging;
      hanging = 0;
    }

    // 11. if contra, double, and if recontra, double yet again:
    if (contra) {
      int add0;
      int add1;
      if (recontra) {
        add0 = 3 * result[0];
        add1 = 3 * result[1];
      } else {
        add0 = result[0];
        add1 = result[1];
      }
      result[0] += add0;
      result[1] += add1;
      teamScore[0] += add0;
      teamScore[1] += add1;
    }

    // 12. ready, return the result:
    return result;
  }

  private int calculateDifference(int rounding0, int rounding1, BelotColor trump) {
    int sum = rounding0 + rounding1;
    int all = 16;
    if (trump == BelotColor.NoTrump || trump == BelotColor.AllTrump) {
      all += 10;
    }
    return all - sum;
  }

  // Betting:
  public synchronized int getPlayerBettingTurn() {
    if (lastBettingPlayer == bettingTurn) {
      // betting finished:
      return -1;
    }
    return bettingTurn;
  }

  public synchronized boolean canBet(BelotColor color, boolean contra) {
    // null color means a pass:
    if (color == null) {
      return true;
    }
    int cmp = trump == null ? 1 : color.compareTo(trump);
    if (cmp > 0) {
      if (contra) {
        return false;
      }
      return true;
    }
    if (cmp == 0) {
      int playerTeam = getTeamOfPlayer(bettingTurn);
      if (playerTeam == bettingTeam) {
        return false;
      }
      // you can state the same as the opponent if you say contra, and it isn't already a recontra:
      return (!this.recontra) && contra;
    }
    return false;
  }

  public synchronized void bet(BelotColor color, boolean contra) {
    if (color != null) {
      this.bettingTeam = getTeamOfPlayer(bettingTurn);
      if (color != this.trump) {
        this.contra = false;
        this.recontra = false;
        contra = false;
      }
      if (this.contra) {
        this.recontra = contra;
      } else {
        this.contra = contra;
        this.recontra = false;
      }
      this.trump = color;
      lastBettingPlayer = bettingTurn;
    } else {
      if (lastBettingPlayer == -1) {
        lastBettingPlayer = bettingTurn;
      }
    }
    bettingTurn = (bettingTurn + 1) % 4;
  }

  // Trump:
  public synchronized BelotColor getTrump() {
    return trump;
  }

  public synchronized boolean isContra() {
    return contra;
  }

  public synchronized boolean isReContra() {
    return recontra;
  }

  // Playing:
  public synchronized int getPlayerPlayingTurn() {
    return playingTurn;
  }

  public static int findHighest(
    BelotCard[] handOfCards,
    BelotColor queryColor,
    BelotColor trump
  ) {
    int highestIndex = -1;
    for (int i = 0; i < handOfCards.length; ++i) {
      if (handOfCards[i] == null) {
        continue;
      }
      if (handOfCards[i].getColor() == queryColor) {
        if (
          (highestIndex == -1) ||
          (handOfCards[highestIndex].compareTo(handOfCards[i].getNumber(), trump) < 0)
        ) {
          highestIndex = i;
        }
      }
    }
    return highestIndex;
  }

  public static boolean satisfiesRaising(
    BelotCard card,
    BelotCard highest,
    BelotColor trump
  ) {
    return card.compareTo(highest.getNumber(), trump) > 0;
  }

  public static boolean canSatisfyRaising(
    BelotCard[] cards,
    BelotCard card,
    BelotCard highest,
    BelotColor requiredColor,
    BelotColor trump
  ) {
    for (int i = 0; i < cards.length; ++i) {
      if (cards[i] != null && cards[i].getColor() == requiredColor) {
        if (cards[i].compareTo(highest.getNumber(), trump) > 0) {
          return true;
        }
      }
    }
    return false;
  }

  private static boolean raisings(
    BelotCard[] cards,
    BelotCard card,
    BelotCard[] handOfCards,
    BelotColor requiredColor,
    BelotColor trump
  ) {
    int highestIndex = findHighest(handOfCards, requiredColor, trump);
    if (satisfiesRaising(card, handOfCards[highestIndex], trump)) {
      return true;
    }
    return !canSatisfyRaising(cards, card, handOfCards[highestIndex], requiredColor, trump);
  }

  public static int findWinnerCardNonTrump(BelotCard[] handOfCards, BelotColor requiredColor, BelotColor trump) {
    return findHighest(handOfCards, requiredColor, trump);
  }

  public static int findWinnerCardTrump(BelotCard[] handOfCards, BelotColor requiredColor, BelotColor trump) {
    int highestNonTrump = findHighest(handOfCards, requiredColor, trump);
    if (trump == requiredColor) {
      return highestNonTrump;
    }
    int highestTrump = findHighest(handOfCards, trump, trump);
    if (highestTrump != -1) {
      return highestTrump;
    }
    return highestNonTrump;
  }

  public static int findFirstFree(BelotCard[] handOfCards) {
    for (int i = 0; i < handOfCards.length; ++i) {
      if (handOfCards[i] == null) {
        return i;
      }
    }
    return -1;
  }

  public static boolean hasColor(BelotCard[] cards, BelotColor queryColor) {
    for (int i = 0; i < cards.length; ++i) {
      if (cards[i] != null && cards[i].getColor() == queryColor) {
        return true;
      }
    }
    return false;
  }

  public synchronized boolean canPlay(int card) {
    BelotCard[] cards = cardsPerPlayer[playingTurn];
    if (cards[card] == null) {
      return false;
    }
    BelotCard firstHandCard = handOfCards[0];
    if (firstHandCard == null) {
      return true;
    }
    BelotColor requiredColor = firstHandCard.getColor();
    if (trump == BelotColor.NoTrump) {
      if (cards[card].getColor() == requiredColor) {
        return true;
      }
      return !hasColor(cards, requiredColor);
    }
    if (trump == BelotColor.AllTrump) {
      if (cards[card].getColor() == requiredColor) {
        return raisings(cards, cards[card], handOfCards, requiredColor, trump);
      }
      return !hasColor(cards, requiredColor);
    }

//     .               .
//                ,    ,
//  . .  .      ( ),     - ( ).
//      -  ,     .

    if (cards[card].getColor() == requiredColor) {
      if (requiredColor == trump) {
        return raisings(cards, cards[card], handOfCards, requiredColor, trump);
      }
      return true;
    }
    if (hasColor(cards, requiredColor)) {
      return false;
    }
    if (trump == requiredColor) {
      // the player must play a trump, but the player doesn't have a trump!
      return true;
    }
    // the player doesn't have the required color, so the player must play a trump
    // if the enemy is taking the hand!
    int highestCard = findWinnerCardTrump(handOfCards, requiredColor, trump);
    int current = findFirstFree(handOfCards);
    if (highestCard % 2 != current % 2) {
      // the enemy is taking the hand!
      // the player MUST play a trump!
      if (hasColor(cards, trump)) {
        // moreover, the player must play a higher trump.
        // if the player doesn't have a higher trump, the player may play any card!
        int highestHandTrump = findHighest(handOfCards, trump, trump);
        if (highestHandTrump == -1) {
          // any trump will do then:
          return cards[card].getColor() == trump;
        }
        if (cards[card].getColor() == trump) {
          // OK if higher than the highest trump:
          if (cards[card].compareTo(handOfCards[highestHandTrump].getNumber(), trump) > 0) {
            return true;
          }
        }
        int highestPlayerTrump = findHighest(cards, trump, trump);
        if (cards[highestPlayerTrump].compareTo(handOfCards[highestHandTrump].getNumber(), trump) > 0) {
          return false;
        }
      }
    }
    return true;
  }

  public static boolean canHaveBelot(BelotCard[] handOfCards, BelotColor trump, BelotCard[] cards, int card) {
    BelotCard belotCard = cards[card];
    int cardNumber = belotCard.getNumber();
    if (cardNumber != BelotCard.NUMBER_QUEEN && cardNumber != BelotCard.NUMBER_KING) {
      return false;
    }
    BelotColor cardColor = belotCard.getColor();
    if (trump != BelotColor.AllTrump && cardColor != trump) {
      return false;
    }
    if (trump == BelotColor.AllTrump && handOfCards[0] != null && handOfCards[3] == null) {
      if (handOfCards[0].getColor() != cardColor) {
        return false;
      }
    }
    return Announces.findCard(
      cardColor,
      ((cardNumber == BelotCard.NUMBER_QUEEN) ?
        BelotCard.NUMBER_KING :
        BelotCard.NUMBER_QUEEN),
      cards);
  }

  private boolean canHaveBelot(int card) {
    BelotCard[] cards = cardsPerPlayer[playingTurn];
    return canHaveBelot(handOfCards, trump, cards, card);
  }

  public synchronized boolean canAnnounce(int what, int card) {
    if (trump == BelotColor.NoTrump) {
      return false;
    }
    if (what == BelotProtocol.AnnounceBelot) {
      return canHaveBelot(card);
    }
    BelotCard[] cards = cardsPerPlayer[playingTurn];
    return announcesPerPlayer[playingTurn].canAdd(what, cards, card);
  }

  public synchronized void announce(int what, int card) {
    if (what == BelotProtocol.AnnounceBelot) {
      ++belotsPerPlayer[playingTurn];
    } else {
      BelotCard[] cards = cardsPerPlayer[playingTurn];
      announcesPerPlayer[playingTurn].add(what, cards, card);
    }
  }

  public synchronized void play(int card) {
    BelotCard[] cards = cardsPerPlayer[playingTurn];
    int toIndex = findFirstFree(handOfCards);
    handOfCards[toIndex] = cards[card];
    cards[card] = null;
    if (toIndex == 3) {
      // the hand is full!

      // 1) find the winner:
      int winnerCardIndex;
      BelotColor requiredColor = handOfCards[0].getColor();
      if (requiredColor == BelotColor.AllTrump || requiredColor == BelotColor.NoTrump) {
        winnerCardIndex = findWinnerCardNonTrump(handOfCards, requiredColor, trump);
      } else {
        winnerCardIndex = findWinnerCardTrump(handOfCards, requiredColor, trump);
      }

      // 2) find the winning player!
      int winnerPlayer = (playingTurn + 1 + winnerCardIndex) % 4;
      lastWinner = winnerPlayer;
      int winnterTeam = getTeamOfPlayer(winnerPlayer);

      // 3) transfer the cards from the hand to the winner:
      for (int i = 0; i < handOfCards.length; ++i) {
        takenCardsPerTeam[winnterTeam].addElement(handOfCards[i]);
        handOfCards[i] = null;
      }

      playingTurn = winnerPlayer;
      ++numberOfHand;
    } else {
      playingTurn = (playingTurn + 1) % 4;
    }
  }

  public synchronized BelotCard[] getPlayerCards(int player) {
    return cardsPerPlayer[player];
  }

  // Utility:
  public static int getTeamOfPlayer(int player) {
    return player % 2;
  }

  public synchronized BelotCard[] getHandOfCards() {
    return handOfCards;
  }

  public static void main(String[] arg) {
    BelotGame game = new BelotGameImpl();
    game.startGame();
    game.startDealing();
    System.out.println(game.getPlayerBettingTurn());
    System.out.println(game.canBet(BelotColor.AllTrump, false));
    game.bet(BelotColor.AllTrump, false);
    for (int i = 0; i < 3; ++i) {
      System.out.println(game.getPlayerBettingTurn());
      System.out.println(game.canBet(null, false));
      game.bet(null, false);
    }
    System.out.println(game.getPlayerBettingTurn());
    game.finishBetting();
    System.out.println("Playing turn: " + game.getPlayerPlayingTurn());
    BelotCard[] firstPlayerCards = game.getPlayerCards(game.getPlayerPlayingTurn());
    for (int i = 0; i < firstPlayerCards.length; ++i) {
      System.out.println(firstPlayerCards[i]);
    }
  }

  public int getHanging() {
    return hanging;
  }

}