package org.belot.server;

import org.belot.protocol.*;
import java.io.*;
import java.net.*;
import java.util.*;

public class ClientCommandThread implements Runnable {

  // common for all clients:
  private ClientWorker clientWorker;
  private GlobalDispatcher globalDispatcher;

  // player info:
  private BelotPlayer belotPlayer;

  // player net objects:
  private ClientSenderThread clientSenderThread;
  private Socket socket;
  private LineNumberReader inReader;

  public ClientCommandThread(
    ClientWorker clientWorker,
    GlobalDispatcher globalDispatcher,
    Socket socket
  ) {
    try {
      this.clientWorker = clientWorker;
      this.globalDispatcher = globalDispatcher;
      this.socket = socket;

      this.belotPlayer = new BelotPlayer(socket.getInetAddress().getHostAddress());

      clientWorker.constructClient(this);

      clientSenderThread = new ClientSenderThread(socket.getOutputStream());
      globalDispatcher.addClientSenderThread(clientSenderThread);

      inReader = new LineNumberReader(new InputStreamReader(socket.getInputStream(), "UTF8"));

      new Thread(this).start();
    } catch (Throwable error) {
      cleanUp();
    }
  }

  private synchronized void cleanUp() {
    if (socket == null) {
      return;
    }
    try {
      socket.close();
    } catch (Throwable error) {
      // do nothing.
    }

    if (clientSenderThread != null) {
      globalDispatcher.removeClientSenderThread(clientSenderThread);
      clientSenderThread.closeClientSenderThread();
    }

    clientWorker.deconstructClient(this);

    socket = null;
    clientWorker = null;
    globalDispatcher = null;
    belotPlayer = null;
    clientSenderThread = null;
    socket = null;
    inReader = null;
  }

  public void run() {
    try {
      clientWorker.startClient(this);
      String[] parseResult = new String[2];
      for (;;) {
        synchronized (this) {
          // do nothing.
        }
        String line = inReader.readLine();
        if (line == null) {
          throw new RuntimeException();
        }
        if (line.length() == 0) {
          continue;
        }
        String commandId = null;
        String answer = "OK";
        boolean commandResponse = false;
        try {
          BelotServerProtocol.parse(line, parseResult);
          commandId = parseResult[0];
          BelotServerProtocol.parse(parseResult[1], parseResult);
          String command = parseResult[0];
          int oldGameStage = clientWorker.getStage();
          if (BelotProtocol.StartGame.equals(command)) {
            clientWorker.startGame(this);
            commandResponse = true;
          } else if (BelotProtocol.StartDealing.equals(command)) {
            clientWorker.startDealing(this);
            commandResponse = true;
          } else if (BelotProtocol.ListPlayers.equals(command)) {
            clientWorker.listPlayers(this);
            commandResponse = true;
          } else if (BelotProtocol.Direction.equals(command)) {
            BelotServerProtocol.parse(parseResult[1], parseResult);
            int direction = Integer.parseInt(parseResult[0]);
            clientWorker.direction(this, direction);
            commandResponse = true;
          } else if (BelotProtocol.Name.equals(command)) {
            BelotServerProtocol.parse(parseResult[1], parseResult);
            String name = parseResult[0];
            clientWorker.name(this, name);
            commandResponse = true;
          } else if (BelotProtocol.TrumpChoice.equals(command)) {
            BelotServerProtocol.parse(parseResult[1], parseResult);
            BelotColor color = BelotColor.getBetting(Integer.parseInt(parseResult[0]));
            BelotServerProtocol.parse(parseResult[1], parseResult);
            boolean contra = "true".equals(parseResult[0]);
            clientWorker.trumpChoice(this, color, contra);
            commandResponse = true;
          } else if (BelotProtocol.Play.equals(command)) {
            BelotServerProtocol.parse(parseResult[1], parseResult);
            int index = Integer.parseInt(parseResult[0]);
            clientWorker.play(this, index);
            commandResponse = true;
          } else if (BelotProtocol.PlayBelot.equals(command)) {
            BelotServerProtocol.parse(parseResult[1], parseResult);
            int index = Integer.parseInt(parseResult[0]);
            clientWorker.playBelot(this, index);
            commandResponse = true;
          } else if (BelotProtocol.GetCards.equals(command)) {
            clientWorker.getCards(this);
            commandResponse = true;
          } else if (BelotProtocol.Announce.equals(command)) {
            BelotServerProtocol.parse(parseResult[1], parseResult);
            int what = Integer.parseInt(parseResult[0]);
            BelotServerProtocol.parse(parseResult[1], parseResult);
            int card = Integer.parseInt(parseResult[0]);
            clientWorker.announce(this, what, card);
            commandResponse = true;
          } else if (BelotProtocol.Say.equals(command)) {
            clientWorker.say(this, parseResult[1]);
          }
          if (commandResponse) {
            clientWorker.finishCommand(oldGameStage);
          }
          if (BelotProtocol.Say.equals(command)) {
            commandResponse = true;
          }
        } catch (Throwable error) {
          commandResponse = false;
          answer = error.getMessage();
        } finally {
          if (commandId == null) {
            commandId = "Unknown";
          }
          String message = BelotServerProtocol.commandResponse(commandId, commandResponse, answer);
          clientSenderThread.depositMessage(message);
          parseResult[0] = null;
          parseResult[1] = null;
        }
      }
    } catch (Throwable error) {
    } finally {
      cleanUp();
    }
  }

  public BelotPlayer getBelotPlayer() {
    return belotPlayer;
  }

  public ClientSenderThread getClientSenderThread() {
    return clientSenderThread;
  }

}
