package org.belot.server;

import org.belot.protocol.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class makes sure that all commands from all clients
 * are executed synchroneously.
 */
public class ClientWorker {

  // common for all players:
  private PlayerManager playerManager;
  private GameManager gameManager;
  private GlobalDispatcher globalDispatcher;

  public ClientWorker(
    PlayerManager playerManager,
    GameManager gameManager,
    GlobalDispatcher globalDispatcher
  ) {
    synchronized (this) {
      this.playerManager = playerManager;
      this.gameManager = gameManager;
      this.globalDispatcher = globalDispatcher;
    }
  }

  // construct, start and deconstruct clients:
  public synchronized void constructClient(ClientCommandThread commandThread) {
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    playerManager.addPlayer(belotPlayer);
    globalDispatcher.broadcast(BelotServerProtocol.playerConnected(belotPlayer));
  }

  public synchronized void startClient(ClientCommandThread commandThread) {
    ClientSenderThread clientSenderThread = commandThread.getClientSenderThread();
    clientSenderThread.depositMessage(BelotServerProtocol.FirstResponse);
    clientSenderThread.depositMessage(BelotServerProtocol.yourId(commandThread.getBelotPlayer()));
    broadcastBettingOrPlaying();
    fillGame();
    if (gameManager.getStage() == GameManager.BETTING || gameManager.getStage() == GameManager.PLAYING) {
      globalDispatcher.broadcast(BelotServerProtocol.trump(gameManager));
    } else {
      broadcastScores();
    }
    broadcastActivePlayersList();
    commandThread.getClientSenderThread().depositMessage(getHandMessage());
    commandThread.getClientSenderThread().depositMessage(getScoresMessage());
  }

  public synchronized void deconstructClient(ClientCommandThread commandThread) {
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    if (belotPlayer != null) {
      globalDispatcher.broadcast(BelotServerProtocol.playerDisconnected(belotPlayer));
      playerManager.removePlayer(belotPlayer);
      broadcastActivePlayersList();
      fillGame();
    }
  }

  // check active players:
  private void fillGame() {
    playerManager.fillGame();
  }

  // check playing:
  private void checkPlaying(ClientCommandThread commandThread) {
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    if (playerManager.getPlayer(belotPlayer.getDirection()) != belotPlayer) {
      throw new RuntimeException("You are not playing now!");
    }
  }

  // broadcasting:
  private void broadcastBettingOrPlaying() {
    if (gameManager.getStage() == GameManager.BETTING) {
      globalDispatcher.broadcast(BelotServerProtocol.betTurn(gameManager.getBettingDirection()));
    } else if (gameManager.getStage() == GameManager.PLAYING) {
      globalDispatcher.broadcast(BelotServerProtocol.playTurn(gameManager.getPlayingDirection()));
    } else if (gameManager.getStage() == GameManager.NO_DEALING) {
      globalDispatcher.broadcast(BelotProtocol.NoDealingAtPresent);
    } else if (gameManager.getStage() == GameManager.NO_GAME) {
      globalDispatcher.broadcast(BelotProtocol.NoGameAtPresent);
    }
  }

  private String getScoresMessage() {
    return BelotServerProtocol.scores(gameManager);
  }

  private String getHandMessage() {
    PreviousHand previousHand = gameManager.getCurrentHand();
    return BelotServerProtocol.currentHand(previousHand.getCards(), previousHand.getFirstPlayer());
  }

  private void broadcastScores() {
    globalDispatcher.broadcast(getScoresMessage());
  }

  private void broadcastMessage(ClientCommandThread commandThread, String line) {
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    globalDispatcher.broadcast(BelotServerProtocol.message(belotPlayer, line));
  }

  private void broadcastHand() {
    globalDispatcher.broadcast(getHandMessage());
  }

  private void broadcastActivePlayersList() {
    globalDispatcher.broadcast(BelotServerProtocol.activePlayers(playerManager));
  }

  // command handling:
  public synchronized void startGame(ClientCommandThread commandThread) throws Exception {
    checkPlaying(commandThread);
    gameManager.startGame();
    globalDispatcher.broadcast(BelotProtocol.GameStarted);
  }

  public synchronized void startDealing(ClientCommandThread commandThread) throws Exception {
    checkPlaying(commandThread);
    gameManager.startDealing();
    globalDispatcher.broadcast(BelotProtocol.DealingStarted);
  }

  public synchronized void listPlayers(ClientCommandThread commandThread) {
    commandThread.getClientSenderThread().depositMessage(
      BelotServerProtocol.listPlayers(playerManager.getPlayers()));
  }

  public synchronized void direction(ClientCommandThread commandThread, int direction) {
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    belotPlayer.setDirection(direction);
    playerManager.changePlayer(belotPlayer);
    globalDispatcher.broadcast(BelotServerProtocol.playerChanged(belotPlayer));
    broadcastActivePlayersList();
    fillGame();
  }

  public synchronized void name(ClientCommandThread commandThread, String name) {
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    belotPlayer.setName(name);
    globalDispatcher.broadcast(BelotServerProtocol.playerChanged(belotPlayer));
    broadcastActivePlayersList();
  }

  public synchronized void trumpChoice(ClientCommandThread commandThread, BelotColor color, boolean contra) {
    checkPlaying(commandThread);
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    boolean wasContra = gameManager.isContra();
    gameManager.bet(belotPlayer.getDirection(), color, contra);
    if (gameManager.getStage() == GameManager.BETTING || gameManager.getStage() == GameManager.PLAYING) {
      globalDispatcher.broadcast(BelotServerProtocol.trump(gameManager));
    }
    boolean isRealContra = contra && (!wasContra);
    boolean isRealReContra = contra && wasContra;
    globalDispatcher.broadcast(BelotServerProtocol.bet(belotPlayer, color, isRealContra, isRealReContra));
  }

  public synchronized void play(ClientCommandThread commandThread, int index) {
    checkPlaying(commandThread);
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    BelotCard[] cards = gameManager.getPlayerCards(belotPlayer.getDirection());
    BelotCard playedCard = cards[index];
    gameManager.play(belotPlayer.getDirection(), index);
    globalDispatcher.broadcast(BelotServerProtocol.play(belotPlayer, playedCard));
    broadcastHand();
  }

  public synchronized void playBelot(ClientCommandThread commandThread, int index) {
    checkPlaying(commandThread);
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    BelotCard[] cards = gameManager.getPlayerCards(belotPlayer.getDirection());
    BelotCard playedCard = cards[index];
    gameManager.playBelot(belotPlayer.getDirection(), index);
    globalDispatcher.broadcast(BelotServerProtocol.playBelot(belotPlayer, playedCard));
    broadcastHand();
  }

  public synchronized void getCards(ClientCommandThread commandThread) throws Exception {
    checkPlaying(commandThread);
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    BelotCard[] cards = gameManager.getPlayerCards(belotPlayer.getDirection());
    commandThread.getClientSenderThread().depositMessage(
      BelotServerProtocol.listCards(cards));
  }

  public synchronized void announce(ClientCommandThread commandThread, int what, int card) {
    checkPlaying(commandThread);
    BelotPlayer belotPlayer = commandThread.getBelotPlayer();
    gameManager.announce(belotPlayer.getDirection(), what, card);
    globalDispatcher.broadcast(BelotServerProtocol.playerAnnounce(belotPlayer, what));
  }

  public synchronized void say(ClientCommandThread commandThread, String line) {
    broadcastMessage(commandThread, line);
  }

  public synchronized void finishCommand(int oldGameStage) {
    broadcastBettingOrPlaying();
    if (oldGameStage == GameManager.PLAYING && gameManager.getStage() != GameManager.PLAYING) {
      broadcastScores();
    }
  }

  public synchronized int getStage() {
    return gameManager.getStage();
  }

}