package org.belot.server;

import org.belot.protocol.*;

public class GameManagerImpl implements GameManager {

  private BelotGame belotGame;
  private int stage;
  private int[] lastDealingScore;
  private int lastBettingPlayer = -1;
  private PreviousHand previousHand;

  public GameManagerImpl(BelotGame belotGame) {
    synchronized (this) {
      this.belotGame = belotGame;
      this.stage = NO_GAME;
      lastDealingScore = new int[3];
      this.previousHand = new PreviousHand();
    }
  }

  public synchronized int getStage() {
    return stage;
  }

  // always:
  public synchronized int[] getTeamScores() {
    return new int[] {belotGame.getTeamScore(0), belotGame.getTeamScore(1)};
  }
  public synchronized int[] getDealingScores() {
    return lastDealingScore;
  }
  public synchronized PreviousHand getCurrentHand() {
    return previousHand;
  }

  // only in NO_GAME:
  public synchronized void startGame() {
    if (stage != NO_GAME) {
      throw new RuntimeException("You cannot start a game!");
    }
    belotGame.startGame();
    stage = NO_DEALING;
  }

  // only in NO_DEALING:
  public synchronized void startDealing() {
    if (stage != NO_DEALING) {
      throw new RuntimeException("You cannot start a dealing!");
    }
    belotGame.startDealing();
    lastDealingScore[0] = lastDealingScore[1] = lastDealingScore[2] = 0;
    stage = BETTING;
    lastBettingPlayer = -1;
  }

  // only in BETTING:
  public synchronized void bet(int player, BelotColor color, boolean contra) {
    boolean ok = true;
    if (stage != BETTING) {
      ok = false;
    }
    if (ok) {
      if (belotGame.getPlayerBettingTurn() == player && belotGame.canBet(color, contra)) {
        belotGame.bet(color, contra);
        if (color != null) {
          lastBettingPlayer = player;
        }
        if (belotGame.getPlayerBettingTurn() == -1) {
          if (belotGame.isDealingFinished()) {
            // everyone was pass:
            stage = NO_DEALING;
          } else {
            belotGame.finishBetting();
            stage = PLAYING;
          }
        }
      } else {
        ok = false;
      }
    }
    if (!ok) {
      throw new RuntimeException("You cannot bet!");
    }
  }

  public synchronized int getBettingDirection() {
    if (stage != BETTING) {
      throw new RuntimeException("You cannot bet!");
    }
    return belotGame.getPlayerBettingTurn();
  }

  // only in BETTING and PLAYING:
  private void checkTrump() {
    if (stage != BETTING && stage != PLAYING) {
      throw new RuntimeException("You cannot get the trump!");
    }
  }
  public synchronized BelotColor getBet() {
    checkTrump();
    return belotGame.getTrump();
  }
  public synchronized boolean isContra() {
    checkTrump();
    return belotGame.isContra();
  }
  public synchronized boolean isReContra() {
    checkTrump();
    return belotGame.isReContra();
  }
  public synchronized BelotCard[] getPlayerCards(int player) {
    checkTrump();
    return belotGame.getPlayerCards(player);
  }
  public synchronized int getBettingPlayer() {
    checkTrump();
    return lastBettingPlayer;
  }

  // only in PLAYING:
  private void checkPlaying() {
    if (stage != PLAYING) {
      notPlaying();
    }
  }
  private void notPlaying() {
    throw new RuntimeException("You cannot play!");
  }

  public synchronized void play(int player, int index) {
    checkPlaying();
    if (belotGame.getPlayerPlayingTurn() != player) {
      notPlaying();
    }
    if (!belotGame.canPlay(index)) {
      notPlaying();
    }
    BelotCard card = belotGame.getPlayerCards(player)[index];
    belotGame.play(index);
    previousHand.add(card, player);
    System.out.println("PreviousHand = " + previousHand);
    if (belotGame.isDealingFinished()) {
      int[] dealingScore = belotGame.calculateDealingScore();
      lastDealingScore[0] = dealingScore[0];
      lastDealingScore[1] = dealingScore[1];
      lastDealingScore[2] = belotGame.getHanging();
      this.stage = NO_DEALING;
      if (belotGame.isGameFinished()) {
        if (!belotGame.getLastWasNoCardsTaken()) {
          this.stage = NO_GAME;
        }
      }
    }
  }

  public synchronized void playBelot(int player, int index) {
    announceInternal(player, BelotProtocol.AnnounceBelot, index);
    play(player, index);
  }

  public synchronized void announce(int player, int what, int index) {
    if (what == BelotProtocol.AnnounceBelot) {
      notPlaying();
    }
    announceInternal(player, what, index);
  }

  private void announceInternal(int player, int what, int index) {
    checkPlaying();
    if (belotGame.getPlayerPlayingTurn() != player) {
      notPlaying();
    }
    if (what == BelotProtocol.AnnounceBelot && (!belotGame.canPlay(index))) {
      notPlaying();
    }
    if (!belotGame.canAnnounce(what, index)) {
      notPlaying();
    }
    belotGame.announce(what, index);
  }

  public synchronized int getPlayingDirection() {
    checkPlaying();
    return belotGame.getPlayerPlayingTurn();
  }

  // only in NO_GAME and NO_DEALING and PLAYING:
  private void checkAfterHand() {
    if (stage != NO_GAME && stage != NO_DEALING && stage != PLAYING) {
      throw new RuntimeException("You cannot get the hand!");
    }
  }
//  public synchronized boolean isHandFinished() {
//    checkAfterHand();
//    BelotCard[] hand = belotGame.getHandOfCards();
//    return hand[3] != null;
//  }
  public synchronized int getHandWinner() {
    checkAfterHand();
    return belotGame.getPlayerPlayingTurn();
  }

}